#
# VFS for ruby - open anything with URI
# 
#   Copyright: Tatsuki Sugiura <sugi@nemui.org>
#   License: Ruby's original or GPL2
#   $LastChangedDate: 2003-11-12 09:17:13 +0900 (Wed, 12 Nov 2003) $
#   $LastChangedBy: sugi $
#

require 'vfs/file'
require 'vfs/scheme/ftp/common'

module VFS;  module SchemeHandler;  module FTP;
  class File < VFS::File
    include VFS::SchemeHandler::FTP::Common

    def initialize(url)
      super
      @ls_cache   = nil
      @is_fecthed = false
      init_common
    end

    def fetched?
      @is_fecthed
    end

    def fetched=(v)
      @is_fecthed = v
    end

    def rename_internal(dest)
      if dest.host == uri.host &&
         dest.port == uri.port &&
         dest.user == uri.user
        ftp_request { |ftp|
          ftp.rename(uri.path, dest.path)
        }
        @uri = dest
      else
        rename_external(dest)
      end
    end

    def open_for_read(flag = RDONLY, perm = "only ignored")
      fetch
      super
    end

    def open_for_write(flag = WRONLY, perm = nil)
      flag = conv_openflang_to_const(flag)
      if flag & CREAT == 0
        log("no create moge. checking existance.", Log::DEBUG)
        # FIXME. change exception class.
        exists? or raise Errno::ENOENT
      end
      if flag & RDWR != 0 && flag & TRUNC == 0
        log("read-write moge. do fetch.", Log::DEBUG)
        exists? && fetch
      end
      if flag & APPEND != 0
        log("Append moge. set pointer to end of TempBuf.", Log::DEBUG)
        buf.seek(buf.size)
      end

      buf.writeback_proc = Proc.new {
        # for writeback to server
        buf.seek(0)
        cmd = "STOR"
        if flag & APPEND != 0 && exists?
          cmd = "APPE"
        end
        ftp_request { |ftp|
          ftp.storbinary("#{cmd} #{URI.unescape(uri.path).gsub(/[\?\\ ]/, '\\\1')}",
                         buf, Net::FTP::DEFAULT_BLOCKSIZE)
        }
      } # end of close hook for writeback to server

      if block_given?
        begin
          yield(buf)
        ensure
          buf.close
        end
      else
        buf
      end
    end

    def readable?
      # fixme: can't check group readable file...
      if /^.([r-]).....([r-])..\s+\d+\s+(\S+)/.match(ls_ld)
        $2 == "r" ||
            $1 == "r" && (uri.user == nil || $3 == URI.unescape(uri.user))
      else
        false
      end
    end

    def writable?
      uri.user == nil and return false
      # fixme: can't check group writable file...
      if /^..([w-]).....([w-]).\s+\d+\s+(\S+)/.match(ls_ld)
        $2 == "w" || ($1 == "w" && $3 == URI.unescape(uri.user))
      else
        false
      end
    end

    def executable?
      # fixme: can't check group executable file...
      if /^...([x-]).....([x-])\s+\d+\s+(\S+)/.match(ls_ld)
        $2 == "x" ||
            $1 == "x" && (uri.user == nil || $3 == URI.unescape(uri.user))
      else
        false
      end          
    end
    alias_method :executable_real?, :executable?

    def exists?
      begin
        ls_ld
      rescue Net::FTPTempError, Net::FTPPermError, Errno::ENOENT
        false
      end
    end
    alias_method :exist?, :exists?

    def file?
      ls_ld_cmp_at(0, 0x2d) # -
    end

    def directory?
      ls_ld_cmp_at(0, 0x64) # d
    end

    def symlink?
      ls_ld_cmp_at(0, 0x6c) # l
    end

    def readlink
      symlink? or raise # FIXME!!!! proper exception class
      /->\s*(.*)$/.match(ls_ld)[1].strip
    end

    def chardev?
      ls_ld_cmp_at(0, 0x63) # c
    end

    def blockdev?
      ls_ld_cmp_at(0, 0x62) # b
    end

    def pipe?
      ls_ld_cmp_at(0, 0x70) # p
    end

    def socket?
      ls_ld_cmp_at(0, 0x73) # s
    end

    def sticky?
      ls_ld_cmp_at(9, 0x74) # t
    end

    def setuid?
      ls_ld_cmp_at(3, 0x73) || # s
        ls_ld_cmp_at(3, 0x53) # S
    end

    def setgid?
      ls_ld_cmp_at(6, 0x73) || # s
        ls_ld_cmp_at(6, 0x53) # S
    end

    def owned?
      begin
        return false if uri.user == nil
        if /^[rwxSstcblp-]+\s*\d+\s+(\S+)/.match(ls_ld)
          $1 == URI.unescape(uri.user)
        else
          raise Error::Unknown, "may be bug. please report!"
        end
      rescue Errno::ENOENT # fixme proper exception class
        false
      end
    end

    def grpowned?
      return false # fixme? Hmmmmm.....
    end

    def nlink
      if /^[rwxSstcblp-]+\s*(\d+)/.match(ls_ld)
        $1.to_i
      else
        raise Error::Unknown, "may be bug. please report!"
      end
    end

    def size
      begin
        ftp_request { |ftp|
          ftp.size(uri.path) 
        }
      rescue Net::FTPPermError
        if /^[rwxSstcblp-]+\s*\d+\s+\S+\s+\S+\s+(\d+)/.match(ls_ld)
          $1.to_i
        else
          raise # FIXME!!!! proper exception class!
        end
      end
    end

    def mtime
      ftp_request { |ftp|
        ftp.mtime(uri.path) 
      }
    end
    alias_method :ctime, :mtime
    alias_method :atime, :mtime

    def delete
      ftp_request { |ftp|
        ftp.delete(uri.path) 
      }
    end
    alias_method :unlink, :delete

    private
    def ls_ld_cmp_at(pos, cmpchar)
      begin
        ls_ld[pos] == cmpchar
      rescue Errno::ENOENT # fixme proper exception class
        false
      end
    end

    def ls_ld(target = uri.path)
      if @ls_cache == nil || Time.now - @ls_cache[1] > 10 # Hmmm, hard coded..
        @ls_cache =
           [ftp_request { |ftp|
              t = ftp.nlst("-ld #{target.gsub(/([?\\* ])/,'\\\1')}")
              if t.length == 0
                raise Errno::ENOENT
              elsif t.length != 1
                log("weird server response. may be bug", Log::WARNING)
              end
              t[0]
            }, Time.now]
      end
      @ls_cache[0]
    end

    def fetch
      fetched? and return true
      ftp_request { |ftp|
        log("start to fetch file.", Log::DEBUG)
        begin
          ftp.retrbinary("RETR " + uri.path,
                         Net::FTP::DEFAULT_BLOCKSIZE) { |data|
            buf << data
          }
        ensure
          buf.close
        end
      }
      log("file fetching done.", Log::DEBUG)
      __send__ :fetched=, true
    end
  end # VFS::SchemeHandler::FTP::File
end; end; end;
 

if $0 == __FILE__
  uri = "ftp://ftp.ring.gr.jp/index-j.html"
  
  f = VFS::SchemeHandler::FTP::File.new(uri)
  p f.mtime
  p f.size
  p f.exists?
  p f.directory?
  p f.file?
  p f.readable?
  p f.writable?
  p f.nlink
  p f.open.read(20)

  print "Password: "
  system("stty -echo") if STDIN.tty?
  pass = STDIN.gets.chomp
  system("stty echo") if STDIN.tty?
  puts
  uri = "ftp://sugi:#{URI.escape(pass)}@localhost:351/home/sugi/tmp/ftptest"
  w = VFS::SchemeHandler::FTP::File.new(uri)
  w.open("a") { |f|
    File.open("/etc/motd") { |r|
      f << r.read
    }
  }
  w.rename("ftp://sugi:#{URI.escape(pass)}@localhost:351/home/sugi/tmp/ftpmove")
  w.unlink
end
